#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var babelPolyfill = require('babel-polyfill');
var yargs = _interopDefault(require('yargs'));
var fs = _interopDefault(require('fs'));
var zlib = _interopDefault(require('zlib'));
var _ = _interopDefault(require('lodash'));
var DecompressZip = _interopDefault(require('decompress-zip'));
var fse = _interopDefault(require('fs-extra'));
var rimraf = _interopDefault(require('rimraf'));
var crypto = _interopDefault(require('crypto'));
var humanize = _interopDefault(require('string-humanize'));

var asyncGenerator = function () {
  function AwaitValue(value) {
    this.value = value;
  }

  function AsyncGenerator(gen) {
    var front, back;

    function send(key, arg) {
      return new Promise(function (resolve, reject) {
        var request = {
          key: key,
          arg: arg,
          resolve: resolve,
          reject: reject,
          next: null
        };

        if (back) {
          back = back.next = request;
        } else {
          front = back = request;
          resume(key, arg);
        }
      });
    }

    function resume(key, arg) {
      try {
        var result = gen[key](arg);
        var value = result.value;

        if (value instanceof AwaitValue) {
          Promise.resolve(value.value).then(function (arg) {
            resume("next", arg);
          }, function (arg) {
            resume("throw", arg);
          });
        } else {
          settle(result.done ? "return" : "normal", result.value);
        }
      } catch (err) {
        settle("throw", err);
      }
    }

    function settle(type, value) {
      switch (type) {
        case "return":
          front.resolve({
            value: value,
            done: true
          });
          break;

        case "throw":
          front.reject(value);
          break;

        default:
          front.resolve({
            value: value,
            done: false
          });
          break;
      }

      front = front.next;

      if (front) {
        resume(front.key, front.arg);
      } else {
        back = null;
      }
    }

    this._invoke = send;

    if (typeof gen.return !== "function") {
      this.return = undefined;
    }
  }

  if (typeof Symbol === "function" && Symbol.asyncIterator) {
    AsyncGenerator.prototype[Symbol.asyncIterator] = function () {
      return this;
    };
  }

  AsyncGenerator.prototype.next = function (arg) {
    return this._invoke("next", arg);
  };

  AsyncGenerator.prototype.throw = function (arg) {
    return this._invoke("throw", arg);
  };

  AsyncGenerator.prototype.return = function (arg) {
    return this._invoke("return", arg);
  };

  return {
    wrap: function (fn) {
      return function () {
        return new AsyncGenerator(fn.apply(this, arguments));
      };
    },
    await: function (value) {
      return new AwaitValue(value);
    }
  };
}();





var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();







var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var get = function get(object, property, receiver) {
  if (object === null) object = Function.prototype;
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return get(parent, property, receiver);
    }
  } else if ("value" in desc) {
    return desc.value;
  } else {
    var getter = desc.get;

    if (getter === undefined) {
      return undefined;
    }

    return getter.call(receiver);
  }
};

















var set = function set(object, property, value, receiver) {
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent !== null) {
      set(parent, property, value, receiver);
    }
  } else if ("value" in desc && desc.writable) {
    desc.value = value;
  } else {
    var setter = desc.set;

    if (setter !== undefined) {
      setter.call(receiver, value);
    }
  }

  return value;
};

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();











var toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var REPORTS_DESTINATION = '.jsGiven-reports';

var INTRO_WORD_METHODS = ['given', 'when', 'then', 'and', 'but', 'with'];

var ScenarioPart = function () {
    function ScenarioPart(kind) {
        var steps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
        classCallCheck(this, ScenarioPart);

        this.kind = kind;
        this.steps = steps;
        this.introWord = null;
    }

    createClass(ScenarioPart, [{
        key: 'stageMethodCalled',
        value: function stageMethodCalled(methodName, parameters) {
            if (INTRO_WORD_METHODS.find(function (introWord) {
                return introWord === methodName;
            })) {
                this.introWord = methodName;
            } else {
                var isFirstStep = this.steps.length === 0;
                this.steps.push(new Step(methodName, parameters, isFirstStep, this.introWord));
                this.introWord = null;
            }
        }
    }]);
    return ScenarioPart;
}();

var Step = function Step(methodName, parameters, isFirstStep, introWord) {
    classCallCheck(this, Step);

    var TWO_DOLLAR_PLACEHOLDER = 'zzblablaescapedollarsignplaceholdertpolm';

    var parametersCopy = [].concat(toConsumableArray(parameters));
    var words = [].concat(toConsumableArray(methodName // 'a_bill_of_$_$$'
    .replace('$$', TWO_DOLLAR_PLACEHOLDER) // 'a_bill_of_$_TWO_DOLLAR_PLACEHOLDER'
    .split('$') // ['a_bill_of', 'TWO_DOLLAR_PLACEHOLDER']
    .map(function (word) {
        return _.lowerCase(humanize(word));
    }) //  ['a bill of', 'TWO_DOLLAR_PLACEHOLDER']
    .reduce(function (previous, newString, index) {
        if (index === 0) {
            return [newString];
        }

        var formattedParameters = void 0;
        if (parametersCopy.length > 0) {
            var _parametersCopy$splic = parametersCopy.splice(0, 1);

            var _parametersCopy$splic2 = slicedToArray(_parametersCopy$splic, 1);

            var parameter = _parametersCopy$splic2[0];

            formattedParameters = [formatParameter(parameter)];
        } else {
            formattedParameters = [];
        }

        return [].concat(toConsumableArray(previous), toConsumableArray(formattedParameters), [newString]);
    }, []) //  ['a bill of', '500', 'TWO_DOLLAR_PLACEHOLDER']
    .filter(function (word) {
        return word !== '';
    }) // If one puts a $ at the end of the method, this adds a useless '' at the end
    .map(function (word) {
        return word.replace(TWO_DOLLAR_PLACEHOLDER, '$');
    }) //  ['a bill of', '500', '$']
    .map(toWord)), toConsumableArray(parametersCopy.map(formatParameter).map(toWord)));

    if (introWord) {
        words = [toIntroWord(introWord)].concat(toConsumableArray(words));
    }

    if (isFirstStep) {
        var _words = words;

        var _words2 = toArray(_words);

        var _words2$ = _words2[0];
        var value = _words2$.value;
        var isIntroWord = _words2$.isIntroWord;

        var rest = _words2.slice(1);

        words = [new Word(_.upperFirst(value), isIntroWord)].concat(toConsumableArray(rest));
    }
    this.words = words;
    this.name = words.map(function (_ref) {
        var value = _ref.value;
        return value;
    }).join(' ');

    function formatParameter(parameter) {
        if (_.isObject(parameter) || Array.isArray(parameter)) {
            if (parameter.toString && parameter.toString !== Object.prototype.toString && parameter.toString !== Array.prototype.toString) {
                return parameter.toString();
            }
            return JSON.stringify(parameter);
        } else {
            return parameter && parameter.toString ? parameter.toString() : JSON.stringify(parameter);
        }
    }

    function toWord(value) {
        return new Word(value, false);
    }

    function toIntroWord(value) {
        return new Word(value, true);
    }
};

var Word = function Word(value, isIntroWord) {
    classCallCheck(this, Word);

    this.value = value;
    this.isIntroWord = isIntroWord;
};

var ScenarioReport = function () {
    function ScenarioReport(groupReport, name) {
        var parts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
        classCallCheck(this, ScenarioReport);

        this.groupReport = groupReport;
        this.name = name;
        this.parts = parts;
    }

    createClass(ScenarioReport, [{
        key: 'dumpToFile',
        value: function dumpToFile() {
            createDirOrDoNothingIfExists(REPORTS_DESTINATION);
            var fileName = computeScenarioFileName(this.groupReport.name, this.name);
            fs.writeFileSync(REPORTS_DESTINATION + '/' + fileName, JSON.stringify(this), 'utf-8');
        }
    }]);
    return ScenarioReport;
}();

var GroupReport = function GroupReport(name) {
    classCallCheck(this, GroupReport);

    this.name = name;
};

function createDirOrDoNothingIfExists(path) {
    try {
        fs.mkdirSync(REPORTS_DESTINATION);
    } catch (error) {
        if (error.code !== 'EEXIST') {
            throw error;
        } else {
            // do nothing
        }
    }
}

function computeScenarioFileName(groupName, scenarioName) {
    var hash = crypto.createHash('sha256');
    hash.update(groupName + '\n' + scenarioName);
    return hash.digest('hex');
}

function generateJGivenReport() {
    return regeneratorRuntime.async(function generateJGivenReport$(_context) {
        while (1) {
            switch (_context.prev = _context.next) {
                case 0:
                    _context.prev = 0;
                    _context.next = 3;
                    return regeneratorRuntime.awrap(installJGivenReportApp());

                case 3:
                    generateJGivenReportDataFiles();

                    console.log('jsGiven report available in ./jGiven-report');
                    _context.next = 11;
                    break;

                case 7:
                    _context.prev = 7;
                    _context.t0 = _context['catch'](0);

                    console.log(_context.t0);
                    process.exit(-1);

                case 11:
                case 'end':
                    return _context.stop();
            }
        }
    }, null, this, [[0, 7]]);
}

function cleanReports() {
    return regeneratorRuntime.async(function cleanReports$(_context2) {
        while (1) {
            switch (_context2.prev = _context2.next) {
                case 0:
                    _context2.prev = 0;
                    _context2.next = 3;
                    return regeneratorRuntime.awrap(removeDir('./' + REPORTS_DESTINATION));

                case 3:
                    _context2.next = 9;
                    break;

                case 5:
                    _context2.prev = 5;
                    _context2.t0 = _context2['catch'](0);

                    console.log(_context2.t0);
                    process.exit(-1);

                case 9:
                case 'end':
                    return _context2.stop();
            }
        }
    }, null, this, [[0, 5]]);
}

function installJGivenReportApp() {
    var reportPrefix = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '.';
    var reportDir, jarFile;
    return regeneratorRuntime.async(function installJGivenReportApp$(_context3) {
        while (1) {
            switch (_context3.prev = _context3.next) {
                case 0:
                    reportDir = reportPrefix + '/jGiven-report';
                    _context3.next = 3;
                    return regeneratorRuntime.awrap(removeDir(reportDir));

                case 3:
                    fs.mkdirSync(reportDir);
                    fs.mkdirSync(reportDir + '/data');

                    jarFile = './jgiven-html5-report.jar';

                    if (fileExists(jarFile)) {
                        _context3.next = 10;
                        break;
                    }

                    jarFile = 'node_modules/js-given/jgiven-html5-report.jar';

                    if (fileExists(jarFile)) {
                        _context3.next = 10;
                        break;
                    }

                    throw new Error('jgiven html 5 report not found');

                case 10:
                    fse.copySync(jarFile, reportDir + '/jgiven-html5-report.jar');

                    _context3.next = 13;
                    return regeneratorRuntime.awrap(unzip(reportDir + '/jgiven-html5-report.jar', reportDir));

                case 13:
                    _context3.next = 15;
                    return regeneratorRuntime.awrap(unzip(reportDir + '/com/tngtech/jgiven/report/html5/app.zip', reportDir));

                case 15:
                    _context3.next = 17;
                    return regeneratorRuntime.awrap(removeDir(reportDir + '/META-INF'));

                case 17:
                    _context3.next = 19;
                    return regeneratorRuntime.awrap(removeDir(reportDir + '/com'));

                case 19:
                    _context3.next = 21;
                    return regeneratorRuntime.awrap(removeDir(reportDir + '/jgiven-html5-report.jar'));

                case 21:
                case 'end':
                    return _context3.stop();
            }
        }
    }, null, this);
}

function fileExists(fileName) {
    try {
        return fs.statSync(fileName).isFile();
    } catch (error) {
        return false;
    }
}

function generateJGivenReportDataFiles() {
    var filter = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : function () {
        return true;
    };
    var reportPrefix = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '.';

    var files = fs.readdirSync('./' + REPORTS_DESTINATION).filter(filter);
    var scenarioReports = files.map(function (file) {
        return JSON.parse(fs.readFileSync(REPORTS_DESTINATION + '/' + file, 'utf-8'));
    });
    var groupNamesAndScenarios = Object.entries(_.groupBy(scenarioReports, function (_ref) {
        var name = _ref.groupReport.name;
        return name;
    }));
    var scenarioModels = groupNamesAndScenarios.map(function (_ref2) {
        var _ref3 = slicedToArray(_ref2, 2);

        var groupName = _ref3[0];
        var scenarioReports = _ref3[1];
        return toReportModel(groupName, scenarioReports);
    });

    var reportDir = reportPrefix + '/jGiven-report';

    fs.writeFileSync(reportDir + '/data/metaData.js', 'jgivenReport.setMetaData({"created":"' + new Date().toLocaleString() + '","title":"J(s)Given Report","data":["data0.js"]} );\n', 'utf-8');
    fs.writeFileSync(reportDir + '/data/tags.js', 'jgivenReport.setTags({});\n', 'utf-8');

    var json = JSON.stringify({ scenarios: scenarioModels });
    var buffer = zlib.gzipSync(new Buffer(json, 'utf-8'));
    var base64 = buffer.toString('base64');
    fs.writeFileSync(reportDir + '/data/data0.js', 'jgivenReport.addZippedScenarios(\'' + base64 + '\');\n', 'utf-8');
}

function removeDir(dir) {
    return new Promise(function (resolve, reject) {
        rimraf(dir, function (error) {
            if (error) {
                reject(error);
            } else {
                resolve();
            }
        });
    });
}

function unzip(zipFile, targetDirectory) {
    return new Promise(function (resolve, reject) {
        var unzipper = new DecompressZip(zipFile);

        unzipper.on('error', function (err) {
            console.log('Caught an error when extracting: ' + zipFile + ' to ' + targetDirectory);
            reject(err);
        });

        unzipper.on('extract', function (log) {
            resolve();
        });

        unzipper.extract({
            path: targetDirectory
        });
    });
}

function toReportModel(groupName, scenarioReports) {
    return {
        className: groupName,
        description: groupName,
        name: groupName,
        scenarios: scenarioReports.map(toScenarioModel),
        tagMap: {}
    };
}

function toScenarioModel(scenarioReport, index) {
    return {
        testMethodName: scenarioReport.name,
        tagIds: [],
        scenarioCases: [{
            caseNr: 0,
            derivedArguments: [],
            description: '',
            durationInNanos: 42,
            errorMessage: null,
            explicitParameters: [],
            explicitArguments: [],
            stackTrace: null,
            success: true,
            steps: _.flatMap(scenarioReport.parts, toSteps)
        }],
        casesAsTable: false,
        className: scenarioReport.groupReport.name,
        derivedParameters: [],
        description: scenarioReport.name,
        durationInNanos: 42,
        executionStatus: 'SUCCESS',
        explicitParameters: [],
        extendedDescription: ''
    };
}

function toSteps(scenarioPart) {
    return scenarioPart.steps.map(function (step) {
        return {
            attachment: null,
            durationInNanos: 42,
            extendedDescription: '',
            isSectionTitle: false,
            name: step.methodName,
            nestedSteps: [],
            status: 'PASSED',
            words: step.words.map(function (word) {
                return _extends({}, word.isIntroWord ? { isIntroWord: true } : {}, {
                    value: word.value
                });
            })
        };
    });
}

var _yargs$usage$command$ = slicedToArray(yargs.usage('Usage: $0 <command>').command('report', 'Generate the reports').command('clean', 'Remove the reports and intermediary files').demandCommand(1).argv._, 1);

var command = _yargs$usage$command$[0];


if (command === 'report') {
    generateJGivenReport();
}

if (command === 'clean') {
    cleanReports();
}
